/*
 ********************************************************************
 ********************************************************************
 * RTP-Speak.                                                       *
 * Versione 1.0 (Maggio 2001)                                       * 
 * Principale routine di trasmissione.                              *
 * Gestisce:                                                        *
 * 1. L'acquisizione da microfono                                   *
 * 2. La codifica                                                   *
 * 3. La trasmissione                                               *
 ********************************************************************   
 ********************************************************************
 * Autore: Daniele Mari                                             *
 * e-mail: daniele.mari@tiscalinet.it                               *
 ********************************************************************
 ********************************************************************
                                                                    */


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>	   		// bzero()
#include <sys/soundcard.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>	   			// signal()
#include <arpa/inet.h>	   		// inet_pton()
#include <sys/socket.h>    		// struct sockadrr 
#include <netinet/in.h>    		// sockaddr_in 
#include <netdb.h>	   			// gethostbyname()
#include <stdlib.h>	   			// perror() 
#include "rtp.h"	   			// RTP headers 


#define  MAXLEN 350         /*
 						 	 * massimo num di byte associabili ad 
                             * un pacchetto RTP (  consigliato non 
                             * superare questa soglia) 
                             */

static int loop = 0;


/*
 * Dichiarazioni globali per l'acqusizione audio
 *
 */

static char au_name[15] = "/dev/dsp";

int dsp_stereo = 0; 		// mono
int dsp_speed = 8000;  		// PCMA (defoult), PCMU, G729
extern int au_fd;
char *audio_buff;
int dsp_speed, aubuf_size;

typedef struct {
 char *encoder;
 int  rate;
 int  channel;
} enc_map;
static enc_map pt_map[128];	  

static int sockfd[2];    		// output sockets 

/*
 * La funzione che segue stampa un breve help. 
 */
static void usage (char *processname)
{
 
 fprintf(stderr,"\n");
 fprintf(stderr,"---------------------------------------------------\n");
 fprintf(stderr,"******************** RTP-Speak ********************\n");
 fprintf(stderr,"************** Version 1.0, May 2001 **************\n");
 fprintf(stderr,"\n");
 fprintf(stderr,">>>>>>>>>>>> Designed by: Daniele Mari <<<<<<<<<<<<\n");
 fprintf(stderr,"---------------------------------------------------\n");
 fprintf(stderr,"\n");
 fprintf ( stderr,
         "Usage: %s [-d devicename][-c codername] destination\n",   
          processname);
 fprintf(stderr,"\t-d\t\tdevice name\n");
 fprintf(stderr,"\t-c\t\ttype of encoding name\n");
 fprintf(stderr,
	        "\t\t\tAvailables coders are: PCMU, PCMA, G729\n");
 fprintf(stderr,
         "\tdestination\tIP address or host_name to contact\n");
 fprintf(stderr, "\t escape character is : ^C\n"); 

 exit(1);

} /* usage */



/*
* La funzione che segue verifica la presenza del tipo di codificatore     
* nella 'pt-map'. 
* In caso di successo restituisce il PT corrispondente, altrimenti -1.
*/
static int search( enc_map *ptr, char *name)
{
 register int i;
 
 for(i=0;i<128;i++)
  if( !strcmp(ptr[i].encoder, name) ) return i;
 return -1;

} /* search */


/* 
 * Qesta funzione inizializza la struttura (IPv4 sockaddr) del
 * destinatario con l'indirizzo IP (fornito da riga di comando)ed il
 * numero di porta (di defoult, 5004) di quest'ultimo.
 * L'indirizzo IP pu essere sia in formato dotted-decimal 
 * (151.100.9.29)che in formato hostname (pcvoce2.ing.uniroma1.it).
 */
static void set_addr( char *h, struct sockaddr *sa)
{

 struct hostent *hptr;
 struct sockaddr_in *servaddr=(struct sockaddr_in*)sa; 

 // tipica inizializzazione a 0 della struttura 'sockaddr' 
 bzero(servaddr,sizeof(*servaddr));

 servaddr->sin_family = AF_INET;
 servaddr->sin_port = htons(5004);

 // Se l'IPv4  in formato hostname, lo si riconverte via DSN. 
 if( (hptr = gethostbyname(h)) ) {
   servaddr->sin_addr = *(struct in_addr*)(hptr->h_addr_list[0]);
 }
 // Verifica se l'IPv4  in formato dotted-decimal 
 else if( (inet_pton(AF_INET,h,&servaddr->sin_addr)) <=0 ) {
   fprintf(stderr,
	   "inet_pton error for %s. Maybe destination have no valid format\n", h);
   exit(1);
 } 
} /* set_addr */


/*
 * Funzione che genera l'intestazione RTP. 
 *
 * La variabile 'packet'  di output e conterr 12 byte di 
 * intestazione RTP, ai quali, seguir il payload fornito dal 
 * codificatore.
 *
 * La variabile 'map_position'  di input e serve ad indicare il 
 * valore del payload-type.
 *
 * (N.B. IL PUNTATORE 'h' E' TALE DA CONSERVARE IL VALORE ASSEGNATOGLI
 * TRA UNA CHIAMATA E LA SUCCESSIVA ---> poich, puntando
 * all'indirizzo di 'packet', il valore precedentemente assegnatogli 
 * viene mantenuto).
 * 
 * La variabile 'loop'  di input e serve a sapere se si st trattando
 * il primo pacchetto o i successivi.
 *
 * La variabile 'numBp'  di input e serve ad indicare la lunghezza,
 * in byte, del payload in base al tipo di codifica.
 * 
 * La variabile 'padding'  di input e serve ad indicare quanti byte 
 * di padding sono presenti nel pacchetto.
 */
void rtp(char *packet, int map_position, int loop, int numBp, int padding)
{
 
 rtp_hdr_t *h = (rtp_hdr_t *)packet;

 /* 
  * campi fissi
  */
  
 // setup version 
 h->version = RTP_VERSION;

 // setup padding-bit 
 if( !padding ) 
  h->p = 0;
 else
  h->p = 1;

 // setup header extension 
 h->x=0;

 // setup number of CSRC 
 h->cc=0;

 // setup marker-bit 
 h->m=0;

 // setup payload type 
 h->pt = map_position;
  
 /* 
  * campi variabili
  */
 if( loop == 1 ) {
	
  // setup sequence number	
  h->seq=htons( rand() );    // s sta per 16bit 

  // setup timestamp 
  h->ts=htonl( rand() );     // l sta per 32bit 

  // setup syncronization source (SSRC) identifier 
  h->ssrc=htonl( rand() );
  
 }
 else { // increments variable fields
	
  h->seq = htons( ntohs(h->seq) + 1 ); 
    
  h->ts = htonl( ntohl(h->ts) + numBp );
  
 }
  
 // contributing source (CSRC) supposti assenti 
   
} /* rtp */


/*
 * Funzione di fine.
 * Stampa un breve messaggio di notifica.
 * E' innescata, in modo asincrono, dall'utente mediante 
 * il comando: ^C
 */
static void stop(int signo)
{
 
 fprintf(stderr,"Trasmission interrupted\n");
 fprintf(stderr,"%d packets sent\n",loop); 
 fprintf(stderr, "Bye :-)\n");
 fprintf(stderr, "\n");
 
 exit(0);
 
} /* stop */



// main function
int main (int argc, char *argv[])
{
 static struct sockaddr_in servaddr;
 extern char *optarg;
 extern int optind,opterr,optopt;
 int map_position; 					// Posizione del codificatore nella 'pt_map'.
 char packet[MAXLEN];
 int p,c,i;
 const int on;
 int padding = 0; 					// # di padding-byte in un pacchetto
 int numBp = 160; 					// # byte per pacchetto (defoult: PCMA,20ms)
 char cod_name[5]="PCMA"; 			// defoult: A-law (Europe)
 extern void set_snd(char *audio_name, int open_mod, int dsp_stereo);
 extern int g711(char type_of_conv[2], char *data_in, char *data_out,
		 int padding); 

 // Elabora i comandi forniti da prompt.
 while( (c = getopt(argc, argv, "d:c:h")) != -1 ) {
  switch(c) {
  case 'd':
    strncpy(au_name, optarg, 15);
    break;
  case 'c':
    strncpy(cod_name, optarg, 5); 
    break;
  case '?':
  case 'h': 			// h sta per help
    usage(argv[0]); 
    break;
  }
 }
 
 fprintf(stderr,"\n");
 fprintf(stderr,"---------------------------------------------------\n");
 fprintf(stderr,"******************** RTP-Speak ********************\n");
 fprintf(stderr,"************** Version 1.0, May 2001 **************\n");
 fprintf(stderr,"\n");
 fprintf(stderr,">>>>>>>>>>>> Designed by: Daniele Mari <<<<<<<<<<<<\n");
 fprintf(stderr,"---------------------------------------------------\n");
 fprintf(stderr,"\n");
 fprintf(stderr, "escape character is : ^C\n");
 fprintf(stderr, "\n");

  
 if(optind == argc) {
  fprintf(stderr,"Error: destination (IP address or hostname) missing\n");
  usage(argv[0]);
 }
 else {
	
  // Si inizializza l'indirizzo sockaddr a cui inviare i dati
  set_addr(argv[optind], (struct sockaddr *)&servaddr);

  // si creano due socket UDP-nonconnesse per l'RTP e l'RTCP
  //for(i=0;i<2;i++) {
   sockfd[0] = socket(AF_INET, SOCK_DGRAM, 0);
   if( sockfd[0] < 0 ) {
    perror("socket");
    exit(1);
   }
  // servaddr.sin_port = htons( ntohs(servaddr.sin_port) +i );
  //}
  
 } /* else */
 
 // Messa in piedi dalla lista payload type  
 for (i=0; i<128; i++) {
  pt_map[i].encoder = "xxxx"; // IMPORTANTE [Schildt,136] 
  pt_map[i].rate    = 0;
  pt_map[i].channel = 0;
 }

 pt_map[  0].encoder="PCMU"; pt_map[  0].rate=8000; pt_map[  0].channel=1;
 pt_map[  8].encoder="PCMA"; pt_map[  8].rate=8000; pt_map[  0].channel=1;
 pt_map[ 18].encoder="G729"; pt_map[ 18].rate=8000; pt_map[ 18].channel=1;
  
 i = search(pt_map, cod_name);
 if( i == -1 ) {
  fprintf(stderr,"Connection impossible: %s unavailable.\n",cod_name);
  fprintf(stderr,"Availables coders are: PCMU, PCMA, G729\n");
  exit(1);
 }
 else map_position = i;  
   
/*
 *********************** PERCORSO DI TRASMISSIONE DI ******************* 
 ************************ OGNI SINGOLO PACCHETTO *********************** 
																	   */
																	 
 if( setsockopt( sockfd[0], SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) < 0 )
  perror("SO_BROADCAST");
 // Processo d'acqisizione, da microfono, dei dati audio
 set_snd(au_name, O_RDONLY, dsp_stereo);
 
 signal(SIGINT, stop);
 
 // Elaborazione dei pacchetti
 for( ; ; ) {
  
 /*
  * Leggi la voce registrata. 
  * PCM Lineare, 16 bits/campione ---> 20ms = 320 byte
  */
  if( ( p = read(au_fd, audio_buff,320)) == -1 ) { 
     perror(au_name);
     exit(1);
    }
  if( p < 320 )
   padding = 320 - p;
  
  loop ++;
	
  // Elabora i dati in base al tipo di codifica.
  switch( map_position ) { 
  case 0: // PCMU
  case 8: // PCMA 

    // insert RTP-header
    rtp(packet, map_position, loop, numBp, padding);   
  
    // Inizia le appropriate routine di codifica
    if( map_position == 8 )	
     g711("la", audio_buff, packet+12, padding);
    else 
     g711("lu", audio_buff, packet+12, padding);

    break;
  case 18:  // G.729
   numBp = 20;
   
    fprintf(stderr, 
	    "Sorry, %s coder is unavailable at moment\n",
	    pt_map[18].encoder);
	    
    exit(0);

    break;
  }
 
 // Spedisci l'attuale pacchetto
 if( sendto(sockfd[0], packet, numBp+12, 
	    0, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1 ) 
  perror("sendto");  
 
 } /* for */  
 
 
 return 0;

} /* main */

